! WHIZARD 2.2.0 alpha Sep 12 2013
! 
! Copyright (C) 1999-2013 by 
!     Wolfgang Kilian <kilian@physik.uni-siegen.de>
!     Thorsten Ohl <ohl@physik.uni-wuerzburg.de>
!     Juergen Reuter <juergen.reuter@desy.de>
!     
!     with contributions by 
!     Christian Speckner <cnspeckn@googlemail.com> 
!     and  Felix Braam, Sebastian Schmidt, Daniel Wiesler 
!
! WHIZARD is free software; you can redistribute it and/or modify it
! under the terms of the GNU General Public License as published by 
! the Free Software Foundation; either version 2, or (at your option)
! any later version.
!
! WHIZARD is distributed in the hope that it will be useful, but
! WITHOUT ANY WARRANTY; without even the implied warranty of
! MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the 
! GNU General Public License for more details.
!
! You should have received a copy of the GNU General Public License
! along with this program; if not, write to the Free Software
! Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
!
!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!
! This file has been stripped of most comments.  For documentation, refer
! to the source 'whizard.nw'

module rt_data

  use kinds, only: default !NODEP!
  use iso_varying_string, string_t => varying_string !NODEP!
  use file_utils !NODEP!
  use system_dependencies !NODEP!
  use diagnostics !NODEP!
!  use tao_random_numbers !NODEP!
  use unit_tests

  use pdf_builtin !NODEP!
  use sf_lhapdf !NODEP!
  use os_interface
  use ifiles
  use lexers
  use parser
  use models
  use flavors
  use variables
  use expressions
  use beams
  use process_libraries
  use prclib_stacks
  use prc_core
!  use beam_polarizations
  use beam_structures
!  use user_files
!  use nlo_setup
  use process_stacks
  use iterations

  implicit none
  private

  public :: rt_data_t
  public :: rt_data_test

  type :: rt_parse_nodes_t
     type(parse_node_t), pointer :: cuts_lexpr => null ()
     type(parse_node_t), pointer :: scale_expr => null ()     
     type(parse_node_t), pointer :: fac_scale_expr => null ()
     type(parse_node_t), pointer :: ren_scale_expr => null ()     
     type(parse_node_t), pointer :: weight_expr => null ()
     type(parse_node_t), pointer :: selection_lexpr => null ()
     type(parse_node_t), pointer :: reweight_expr => null ()
     type(parse_node_t), pointer :: analysis_lexpr => null ()
!     type(parse_node_t), pointer :: histogram_writer => null ()
!     type(parse_node_t), pointer :: plot_writer => null ()
!     type(parse_node_t), pointer :: alpha_qed_expr => null ()
   contains
     procedure :: write => rt_parse_nodes_write
  end type rt_parse_nodes_t
     
  type :: rt_data_t
     type(lexer_t), pointer :: lexer => null ()
     type(var_list_t) :: var_list
     type(iterations_list_t) :: it_list
!     type(iterations_list_t), dimension(:), pointer :: it_list_default
     type(os_data_t) :: os_data
     type(model_t), pointer :: model => null ()
     type(prclib_stack_t) :: prclib_stack
     type(process_library_t), pointer :: prclib => null ()
     type(beam_data_t) :: beam_data
!     type(beam_polarization_t), dimension(:), pointer :: &
!        beam_polarization => null ()
     type(beam_structure_t) :: beam_structure
     type(pdf_builtin_status_t) :: pdf_builtin_status
     type(lhapdf_status_t) :: lhapdf_status
     type(rt_parse_nodes_t) :: pn
     type(process_stack_t) :: process_stack
     type(string_t), dimension(:), allocatable :: sample_fmt
!     type(file_list_t), pointer :: out_files => null ()
     logical :: quit = .false.
     integer :: quit_code = 0
!     integer :: environment = -1
!     integer :: analysis_data_unit = -1
!     type(nlo_setup_list_t) :: nlo_setup_list
   contains
     procedure :: write => rt_data_write
     procedure :: write_vars => rt_data_write_vars
     procedure :: write_libraries => rt_data_write_libraries
     procedure :: write_beams => rt_data_write_beams
     procedure :: write_process_stack => rt_data_write_process_stack
     procedure :: global_init => rt_data_global_init
     procedure :: local_init => rt_data_local_init
     procedure :: init_pointer_variables => rt_data_init_pointer_variables
     procedure :: link => rt_data_link
     procedure :: restore => rt_data_restore
     procedure :: final => rt_data_global_final
     procedure :: add_prclib => rt_data_add_prclib
     procedure :: update_prclib => rt_data_update_prclib
     procedure :: get_helicity_selection => rt_data_get_helicity_selection
  end type rt_data_t


contains

  subroutine rt_parse_nodes_write (object, unit)
    class(rt_parse_nodes_t), intent(in) :: object
    integer, intent(in), optional :: unit
    integer :: u
    u = output_unit (unit)
    call wrt ("Cuts", object%cuts_lexpr)
    call write_separator (u)
    call wrt ("Scale", object%scale_expr)
    call write_separator (u)
    call wrt ("Factorization scale", object%fac_scale_expr)
    call write_separator (u)
    call wrt ("Renormalization scale", object%ren_scale_expr)
    call write_separator (u)
    call wrt ("Weight", object%weight_expr)
    call write_separator_double (u)
    call wrt ("Event selection", object%selection_lexpr)
    call write_separator (u)
    call wrt ("Event reweighting factor", object%reweight_expr)
    call write_separator (u)
    call wrt ("Event analysis", object%analysis_lexpr)
!     call write_separator (u)
!     call wrt ("Histogram writer", object%histogram_writer)
!     call write_separator (u)
!     call wrt ("Plot writer", object%plot_writer)
!     call write_separator (u)
!     call wrt ("Alpha(QED)", object%alpha_qed_lexpr)
  contains
    subroutine wrt (title, pn)
      character(*), intent(in) :: title
      type(parse_node_t), intent(in), pointer :: pn
      if (associated (pn)) then
         write (u, "(1x,A,':')")  title
         call write_separator (u)
         call parse_node_write_rec (pn, u)
      else
         write (u, "(1x,A,':',1x,A)")  title, "[undefined]"
      end if
    end subroutine wrt
  end subroutine rt_parse_nodes_write
    
  subroutine rt_data_write (object, unit, vars)
    class(rt_data_t), intent(in) :: object
    integer, intent(in), optional :: unit
    type(string_t), dimension(:), intent(in), optional :: vars
    integer :: u, i
    u = output_unit (unit)
    call write_separator_double (u)
    write (u, "(1x,A)")  "Runtime data:"
    if (present (vars)) then
       if (size (vars) /= 0) then
          call write_separator_double (u)
          write (u, "(1x,A)")  "Selected variables:"
          call write_separator (u)
          call object%write_vars (u, vars)
       end if
    else
       call write_separator_double (u)
       call var_list_write (object%var_list, u, follow_link=.true.)
    end if
    if (object%it_list%get_n_pass () > 0) then
       call write_separator_double (u)
       write (u, "(1x)", advance="no")
       call object%it_list%write (u)
    end if
    call object%prclib_stack%write (u)
!    call write_separator_double (u)
    call beam_data_write (object%beam_data, u)
    call write_separator_double (u)
    call object%beam_structure%write (u)
    call write_separator_double (u)
    call object%pn%write (u)
    if (allocated (object%sample_fmt)) then
       call write_separator (u)
       write (u, "(1x,A)", advance="no")  "Event sample formats = "
       do i = 1, size (object%sample_fmt)
          if (i > 1)  write (u, "(A,1x)", advance="no")  ","
          write (u, "(A)", advance="no")  char (object%sample_fmt(i))
       end do
       write (u, "(A)")
    end if
!    call write_separator_double (u)
    call object%process_stack%write (u)
  end subroutine rt_data_write
  
  subroutine rt_data_write_vars (object, unit, vars)
    class(rt_data_t), intent(in) :: object
    integer, intent(in), optional :: unit
    type(string_t), dimension(:), intent(in), optional :: vars
    integer :: u, i
    u = output_unit (unit)
    if (present (vars)) then
       do i = 1, size (vars)
          call var_list_write_var (object%var_list, vars(i), unit = u, &
               follow_link = .true.)
       end do
    end if
  end subroutine rt_data_write_vars
  
  subroutine rt_data_write_libraries (object, unit)
    class(rt_data_t), intent(in) :: object
    integer, intent(in), optional :: unit
    integer :: u
    u = output_unit (unit)
    call object%prclib_stack%write (u)
  end subroutine rt_data_write_libraries

  subroutine rt_data_write_beams (object, unit)
    class(rt_data_t), intent(in) :: object
    integer, intent(in), optional :: unit
    integer :: u
    u = output_unit (unit)
    call write_separator_double (u)
    call beam_data_write (object%beam_data, u)
    call write_separator_double (u)
    call object%beam_structure%write (u)
    call write_separator_double (u)
  end subroutine rt_data_write_beams

  subroutine rt_data_write_process_stack (object, unit)
    class(rt_data_t), intent(in) :: object
    integer, intent(in), optional :: unit
    call object%process_stack%write (unit)
  end subroutine rt_data_write_process_stack
  
  subroutine rt_data_global_init (global, paths)
    class(rt_data_t), intent(out), target :: global
    type(paths_t), intent(in), optional :: paths
    logical, target, save :: known = .true.
    real(default), parameter :: real_specimen = 1.
    call os_data_init (global%os_data, paths)
!    allocate (global%out_files)
!    allocate (global%rng)
!    call system_clock (global%seed)
!    call tao_random_create (global%rng, global%seed)
    call var_list_append_log_ptr &
         (global%var_list, var_str ("?logging"), logging, known, &
         intrinsic=.true.)
!    call var_list_append_int_ptr &
!         (global%var_list, var_str ("seed_value"), global%seed, known, &
!          intrinsic=.true.)
    call var_list_append_string &
         (global%var_list, var_str ("$model_name"), &
          intrinsic=.true.)
    call var_list_append_string &
         (global%var_list, var_str ("$method"), var_str ("omega"), &
         intrinsic=.true.)        
    call var_list_append_log &
         (global%var_list, var_str ("?report_progress"), .true., &
          intrinsic=.true.)
    call var_list_append_string &
         (global%var_list, var_str ("$restrictions"), var_str (""), &
         intrinsic=.true.)
    call var_list_append_string &
         (global%var_list, var_str ("$omega_flags"), var_str (""), &
         intrinsic=.true.)
!     call var_list_append_log &
!          (global%var_list, var_str ("?read_color_factors"), .true., &
!           intrinsic=.true.)
!     call var_list_append_string &
!          (global%var_list, var_str ("$user_procs_cut"), var_str (""), &
!           intrinsic=.true.)     
!     call var_list_append_string &
!          (global%var_list, var_str ("$user_procs_event_shape"), var_str (""), &
!           intrinsic=.true.)     
!     call var_list_append_string &
!          (global%var_list, var_str ("$user_procs_obs1"), var_str (""), &
!           intrinsic=.true.)     
!     call var_list_append_string &
!          (global%var_list, var_str ("$user_procs_obs2"), var_str (""), &
!           intrinsic=.true.)     
!     call var_list_append_string &
!          (global%var_list, var_str ("$user_procs_sf"), var_str (""), &
!           intrinsic=.true.)     
!     call var_list_append_log &
!          (global%var_list, var_str ("?slha_read_input"), .true., &
!           intrinsic=.true.)
!     call var_list_append_log &
!          (global%var_list, var_str ("?slha_read_spectrum"), .true., &
!           intrinsic=.true.)
!     call var_list_append_log &
!          (global%var_list, var_str ("?slha_read_decays"), .false., &
!           intrinsic=.true.)
    call var_list_append_string &
         (global%var_list, var_str ("$library_name"), &
          intrinsic=.true.)
    call var_list_append_real &
         (global%var_list, var_str ("sqrts"), &
          intrinsic=.true.)
    call var_list_append_real &
         (global%var_list, var_str ("beam1_momentum"), &
          intrinsic=.true.)
    call var_list_append_real &
         (global%var_list, var_str ("beam2_momentum"), &
          intrinsic=.true.)
    call var_list_append_real &
         (global%var_list, var_str ("crossing_angle"), &
          intrinsic=.true.)
    call var_list_append_real &
         (global%var_list, var_str ("beams_theta"), &
          intrinsic=.true.)
    call var_list_append_real &
         (global%var_list, var_str ("beams_phi"), &
          intrinsic=.true.)
    call var_list_append_real &
         (global%var_list, var_str ("luminosity"), 0._default, &
          intrinsic=.true.)
    if (present (paths)) then
       call var_list_append_string &
            (global%var_list, var_str ("$lhapdf_dir"), paths%lhapdfdir, &
             intrinsic=.true.)
    else
       call var_list_append_string &
            (global%var_list, var_str ("$lhapdf_dir"), var_str(""), &
             intrinsic=.true.)
    end if 
    call var_list_append_string &
         (global%var_list, var_str ("$lhapdf_file"), var_str (""), &
          intrinsic=.true.)
    call var_list_append_int &
         (global%var_list, var_str ("lhapdf_member"), 0, &
          intrinsic=.true.)
    call var_list_append_int &
         (global%var_list, var_str ("lhapdf_photon_scheme"), 0, &
          intrinsic=.true.)
    call var_list_append_real &
         (global%var_list, var_str ("isr_alpha"), 0._default, &
          intrinsic=.true.)
    call var_list_append_real &
         (global%var_list, var_str ("isr_q_max"), 0._default, &
          intrinsic=.true.)
    call var_list_append_real &
         (global%var_list, var_str ("isr_mass"), 0._default, &
          intrinsic=.true.)
    call var_list_append_int &
         (global%var_list, var_str ("isr_order"), 3, &
          intrinsic=.true.)
    call var_list_append_log &
         (global%var_list, var_str ("?isr_recoil"), .false., &
          intrinsic=.true.)
    call var_list_append_real &
         (global%var_list, var_str ("epa_alpha"), 0._default, &
          intrinsic=.true.)
    call var_list_append_real &
         (global%var_list, var_str ("epa_x_min"), 0._default, &
          intrinsic=.true.)
    call var_list_append_real &
         (global%var_list, var_str ("epa_q_min"), 0._default, &
          intrinsic=.true.)
    call var_list_append_real &
         (global%var_list, var_str ("epa_e_max"), 0._default, &
          intrinsic=.true.)
    call var_list_append_real &
         (global%var_list, var_str ("epa_mass"), 0._default, &
          intrinsic=.true.)
    call var_list_append_log &
         (global%var_list, var_str ("?epa_recoil"), .false., &
          intrinsic=.true.)
    call var_list_append_real &
         (global%var_list, var_str ("ewa_x_min"), 0._default, &
          intrinsic=.true.)
    call var_list_append_real &
         (global%var_list, var_str ("ewa_q_min"), 0._default, &
          intrinsic=.true.)
    call var_list_append_real &
         (global%var_list, var_str ("ewa_pt_max"), 0._default, &
          intrinsic=.true.)
    call var_list_append_real &
         (global%var_list, var_str ("ewa_mass"), 0._default, &
          intrinsic=.true.)
    call var_list_append_log &
         (global%var_list, var_str ("?ewa_keep_momentum"), .false., &
          intrinsic=.true.)       
    call var_list_append_log &
         (global%var_list, var_str ("?ewa_keep_energy"), .false., &
          intrinsic=.true.)               
!     call var_list_append_log &
!          (global%var_list, var_str ("?circe1_photon1"), .false., &
!           intrinsic=.true.)   
!     call var_list_append_log &
!          (global%var_list, var_str ("?circe1_photon2"), .false., &
!           intrinsic=.true.)   
!     call var_list_append_real &
!          (global%var_list, var_str ("circe1_sqrts"), &
!           intrinsic=.true.)             
!     call var_list_append_log &
!          (global%var_list, var_str ("?circe1_generate"), .true., &
!           intrinsic=.true.)             
!     call var_list_append_log &
!          (global%var_list, var_str ("?circe1_map"), .true., &
!           intrinsic=.true.)   
!     call var_list_append_int &
!          (global%var_list, var_str ("circe1_ver"), 0, intrinsic=.true.)
!     call var_list_append_int &
!          (global%var_list, var_str ("circe1_rev"), 0, intrinsic=.true.)
!     call var_list_append_int &
!          (global%var_list, var_str ("circe1_acc"), 1, intrinsic=.true.)
!     call var_list_append_int &
!          (global%var_list, var_str ("circe1_chat"), 0, intrinsic=.true.)
!     call var_list_append_real &
!          (global%var_list, var_str ("circe2_sqrts"), &
!           intrinsic=.true.)   
!     call var_list_append_log &
!          (global%var_list, var_str ("?circe2_generate"), .true., &
!           intrinsic=.true.)      
!     call var_list_append_log &
!          (global%var_list, var_str ("?circe2_map"), .true., &
!           intrinsic=.true.)             
!     call var_list_append_log &
!          (global%var_list, var_str ("?circe2_polarized"), .true., &
!           intrinsic=.true.)             
!     call var_list_append_string &    
!          (global%var_list, var_str ("$circe2_file"), &
!           intrinsic=.true.)    
!     call var_list_append_string &    
!          (global%var_list, var_str ("$circe2_design"), var_str ("*"), &
!           intrinsic=.true.)             
!     call var_list_append_string &    
!          (global%var_list, var_str ("$beam_events_file"), &
!           intrinsic=.true.)    
!     call var_list_append_log &
!          (global%var_list, var_str ("?beam_events_warn_eof"), .true., &
!           intrinsic=.true.)             
!     call var_list_append_log &
!          (global%var_list, var_str ("?alpha_s_is_fixed"), .true., &
!           intrinsic=.true.)
!     call var_list_append_log &
!          (global%var_list, var_str ("?alpha_s_from_lhapdf"), .false., &
!           intrinsic=.true.)
!     call var_list_append_log &
!          (global%var_list, var_str ("?alpha_s_from_pdf_builtin"), .false., &
!           intrinsic=.true.)
!     call var_list_append_int &
!          (global%var_list, var_str ("alpha_s_order"), 0, &
!           intrinsic=.true.)
!     call var_list_append_int &
!          (global%var_list, var_str ("alpha_s_nf"), 5, &
!           intrinsic=.true.)
!     call var_list_append_log &
!          (global%var_list, var_str ("?alpha_s_from_mz"), .true., &
!           intrinsic=.true.)
!     call var_list_append_real &
!          (global%var_list, var_str ("lambda_qcd"), 200.e-3_default, &
!           intrinsic=.true.)
!     call var_list_append_log &
!          (global%var_list, var_str ("?fatal_beam_decay"), .true., &
!           intrinsic=.true.)          
    call var_list_append_log &
         (global%var_list, var_str ("?helicity_selection_active"), .true., &
          intrinsic=.true.)
    call var_list_append_real &
         (global%var_list, var_str ("helicity_selection_threshold"), &
          1E10_default, &
          intrinsic=.true.)
    call var_list_append_int &
         (global%var_list, var_str ("helicity_selection_cutoff"), 1000, &
          intrinsic=.true.)
    call var_list_append_string &
         (global%var_list, var_str ("$rng_method"), var_str ("tao"), &
          intrinsic=.true.)
    call var_list_append_string &
         (global%var_list, var_str ("$integration_method"), var_str ("vamp"), &
          intrinsic=.true.)
    call var_list_append_int &
         (global%var_list, var_str ("threshold_calls"), 10, &
          intrinsic=.true.)
    call var_list_append_int &
         (global%var_list, var_str ("min_calls_per_channel"), 10, &
          intrinsic=.true.)
    call var_list_append_int &
         (global%var_list, var_str ("min_calls_per_bin"), 10, &
          intrinsic=.true.)
    call var_list_append_int &
         (global%var_list, var_str ("min_bins"), 3, &
          intrinsic=.true.)
    call var_list_append_int &
         (global%var_list, var_str ("max_bins"), 20, &
          intrinsic=.true.)
    call var_list_append_log &
         (global%var_list, var_str ("?stratified"), .true., &
          intrinsic=.true.)
    call var_list_append_log &
         (global%var_list, var_str ("?use_vamp_equivalences"), .true., &
          intrinsic=.true.)
    call var_list_append_real &
         (global%var_list, var_str ("channel_weights_power"), 0.25_default, &
          intrinsic=.true.)
    call var_list_append_string &
         (global%var_list, var_str ("$phs_method"), var_str ("default"), &
          intrinsic=.true.)
!     call var_list_append_log &
!          (global%var_list, var_str ("?vis_channels"), .false., &
!           intrinsic=.true.)     
!     call var_list_append_log &
!          (global%var_list, var_str ("?check_phs_file"), .true., &
!           intrinsic=.true.)
    call var_list_append_string &
         (global%var_list, var_str ("$phs_file"), var_str (""), &
          intrinsic=.true.)
    call var_list_append_log &
         (global%var_list, var_str ("?phs_only"), .false., &
          intrinsic=.true.)
!     call var_list_append_real &
!          (global%var_list, var_str ("phs_threshold_s"), 50._default, &
!           intrinsic=.true.)
!     call var_list_append_real &
!          (global%var_list, var_str ("phs_threshold_t"), 100._default, &
!           intrinsic=.true.)
!     call var_list_append_int &
!          (global%var_list, var_str ("phs_off_shell"), 2, &
!           intrinsic=.true.)
!     call var_list_append_int &
!          (global%var_list, var_str ("phs_t_channel"), 6, &
!           intrinsic=.true.)
!     call var_list_append_real &
!          (global%var_list, var_str ("phs_e_scale"), 10._default, &
!           intrinsic=.true.)
!     call var_list_append_real &
!          (global%var_list, var_str ("phs_m_scale"), 10._default, &
!           intrinsic=.true.)
!     call var_list_append_real &
!          (global%var_list, var_str ("phs_q_scale"), 10._default, &
!           intrinsic=.true.)
!     call var_list_append_log &
!          (global%var_list, var_str ("?phs_keep_nonresonant"), .true., &
!           intrinsic=.true.)
!     call var_list_append_log &
!          (global%var_list, var_str ("?phs_step_mapping"), .true., &
!           intrinsic=.true.)
!     call var_list_append_log &
!          (global%var_list, var_str ("?phs_step_mapping_exp"), .false., &
!           intrinsic=.true.)
!     call var_list_append_log &
!          (global%var_list, var_str ("?phs_s_mapping"), .false., &
!           intrinsic=.true.)
!     call var_list_append_log &
!          (global%var_list, var_str ("?strfun_multichannel"), .false., &
!           intrinsic=.true.)
    call var_list_append_string &
         (global%var_list, var_str ("$run_id"), var_str (""), &
          intrinsic=.true.)
!     call var_list_append_log &
!          (global%var_list, var_str ("?use_best_grid"), .true., &
!           intrinsic=.true.)
!     call var_list_append_log &
!          (global%var_list, var_str ("?check_grid_file"), .true., &
!           intrinsic=.true.)
!     call var_list_append_real &
!          (global%var_list, var_str ("accuracy_goal"), 0._default, &
!           intrinsic=.true.)
!     call var_list_append_real &
!          (global%var_list, var_str ("error_goal"), 0._default, &
!           intrinsic=.true.)
!     call var_list_append_real &
!          (global%var_list, var_str ("relative_error_goal"), 0._default, &
!           intrinsic=.true.)
!     call var_list_append_log &
!          (global%var_list, var_str ("?vis_history"), .false., &
!           intrinsic=.true.)     
!     call var_list_append_log &
!          (global%var_list, var_str ("?isotropic_decay"), .false., &
!           intrinsic=.true.)
!     call var_list_append_log &
!          (global%var_list, var_str ("?diagonal_decay"), .false., &
!           intrinsic=.true.)
!     call var_list_append_log &
!          (global%var_list, var_str ("?check_event_file"), .true., &
!           intrinsic=.true.)
!     call var_list_append_string &
!          (global%var_list, var_str ("$event_file_version"), var_str (""), &
!           intrinsic=.true.)
    call var_list_append_int &
         (global%var_list, var_str ("n_events"), 0, &
          intrinsic=.true.)
!     call var_list_append_log &
!          (global%var_list, var_str ("?allow_decays"), .true., &
!           intrinsic=.true.)     
    call var_list_append_log &
         (global%var_list, var_str ("?unweighted"), .true., &
          intrinsic=.true.)       
!     call var_list_append_string &
!          (global%var_list, var_str ("$event_normalization"), var_str ("auto"),&
!           intrinsic=.true.)     
!     call var_list_append_log &
!          (global%var_list, var_str ("?negative_weights"), .false., &
!           intrinsic=.true.)     
!     call var_list_append_log &
!          (global%var_list, var_str ("?use_num_id"), .false., &
!           intrinsic=.true.)
!     call var_list_append_log &
!          (global%var_list, var_str ("?keep_beams"), .false., &
!           intrinsic=.true.)
!     call var_list_append_log &
!          (global%var_list, var_str ("?update_parameters"), .true., &
!           intrinsic=.true.)     
!     call var_list_append_log &
!          (global%var_list, var_str ("?update_scale"), .false., &
!           intrinsic=.true.)     
!     call var_list_append_log &
!          (global%var_list, var_str ("?update_alpha_s"), .false., &
!           intrinsic=.true.)     
    call var_list_append_log &
         (global%var_list, var_str ("?update_event"), .false., &
          intrinsic=.true.)       
    call var_list_append_log &
         (global%var_list, var_str ("?update_sqme"), .false., &
          intrinsic=.true.)       
    call var_list_append_log &
         (global%var_list, var_str ("?update_weight"), .false., &
          intrinsic=.true.)       
    call var_list_append_string &
         (global%var_list, var_str ("$sample"), var_str (""), &
          intrinsic=.true.)
    call var_list_append_log &
         (global%var_list, var_str ("?read_raw"), .true., &
          intrinsic=.true.)
    call var_list_append_log &
         (global%var_list, var_str ("?write_raw"), .true., &
          intrinsic=.true.)
!     call var_list_append_string &
!          (global%var_list, var_str ("$extension_raw"), var_str ("evx"), &
!          intrinsic=.true.)
!     call var_list_append_string &
!          (global%var_list, var_str ("$extension_default"), var_str ("evt"), &
!          intrinsic=.true.)
!     call var_list_append_string &
!          (global%var_list, var_str ("$extension_debug"), var_str ("debug"), &
!           intrinsic=.true.)
!     call var_list_append_string &
!          (global%var_list, var_str ("$extension_hepevt"), var_str ("hepevt"), &
!           intrinsic=.true.)
!     call var_list_append_string &
!          (global%var_list, var_str ("$extension_ascii_short"), var_str ("short.evt"), &
!           intrinsic=.true.)
!     call var_list_append_string &
!          (global%var_list, var_str ("$extension_ascii_long"), var_str ("long.evt"), &
!           intrinsic=.true.)    
!     call var_list_append_string &
!          (global%var_list, var_str ("$extension_athena"), var_str ("athena.evt"), &
!           intrinsic=.true.) 
!     call var_list_append_string &
!          (global%var_list, var_str ("$extension_mokka"), var_str ("mokka.evt"), &
!           intrinsic=.true.)     
!     call var_list_append_string &
!          (global%var_list, var_str ("$extension_lhef"), var_str ("lhef"), &
!           intrinsic=.true.)
!     call var_list_append_string &
!          (global%var_list, var_str ("$extension_lha"), var_str ("lha"), &
!           intrinsic=.true.)
!     call var_list_append_string &
!          (global%var_list, var_str ("$extension_hepmc"), var_str ("hepmc"), &
!           intrinsic=.true.)
!     call var_list_append_string &
!          (global%var_list, var_str ("$extension_stdhep"), var_str ("stdhep"), &
!           intrinsic=.true.)
!     call var_list_append_string &
!          (global%var_list, var_str ("$extension_stdhep_up"), var_str ("up.stdhep"), &
!           intrinsic=.true.)
!     call var_list_append_string &
!          (global%var_list, var_str ("$extension_hepevt_verbose"), var_str ("hepevt.verb"), &
!           intrinsic=.true.)
!     call var_list_append_string &
!          (global%var_list, var_str ("$extension_lha_verbose"), var_str ("lha.verb"), &
!           intrinsic=.true.)
!     call var_list_append_int (global%var_list, &
!          var_str ("n_bins"), 20, &
!           intrinsic=.true.)
!     call var_list_append_log (global%var_list, &
!          var_str ("?normalize_bins"), .false., &
!           intrinsic=.true.)
!     call var_list_append_string (global%var_list, &
!          var_str ("$obs_label"), var_str (""), &
!           intrinsic=.true.)
!     call var_list_append_string (global%var_list, &
!          var_str ("$obs_unit"), var_str (""), &
!           intrinsic=.true.)
!     call var_list_append_string (global%var_list, &
!          var_str ("$title"), var_str (""), &
!           intrinsic=.true.)
!     call var_list_append_string (global%var_list, &
!          var_str ("$description"), var_str (""), &
!           intrinsic=.true.)
!     call var_list_append_string (global%var_list, &
!          var_str ("$x_label"), var_str (""), &
!           intrinsic=.true.)
!     call var_list_append_string (global%var_list, &
!          var_str ("$y_label"), var_str (""), &
!           intrinsic=.true.)
!     call var_list_append_int &
!          (global%var_list, var_str ("graph_width_mm"), 130, &
!           intrinsic=.true.)
!     call var_list_append_int &
!          (global%var_list, var_str ("graph_height_mm"), 90, &
!           intrinsic=.true.)
!     call var_list_append_log &
!          (global%var_list, var_str ("?y_log"), .false., &
!           intrinsic=.true.)
!     call var_list_append_log &
!          (global%var_list, var_str ("?x_log"), .false., &
!           intrinsic=.true.)
!     call var_list_append_log &
!          (global%var_list, var_str ("?y_log"), .false., &
!           intrinsic=.true.)
!     call var_list_append_real &
!          (global%var_list, var_str ("x_min"),  &
!           intrinsic=.true.)
!     call var_list_append_real &
!          (global%var_list, var_str ("x_max"),  &
!           intrinsic=.true.)
!     call var_list_append_real &
!          (global%var_list, var_str ("y_min"),  &
!           intrinsic=.true.)
!     call var_list_append_real &
!          (global%var_list, var_str ("y_max"),  &
!           intrinsic=.true.)
!     call var_list_append_string &
!          (global%var_list, var_str ("$gmlcode_bg"), var_str (""), &
!           intrinsic=.true.)
!     call var_list_append_string &
!          (global%var_list, var_str ("$gmlcode_fg"), var_str (""), &
!           intrinsic=.true.)
!     call var_list_append_log &
!          (global%var_list, var_str ("?draw_histogram"), &
!           intrinsic=.true.)
!     call var_list_append_log &
!          (global%var_list, var_str ("?draw_base"), &
!           intrinsic=.true.)
!     call var_list_append_log &
!          (global%var_list, var_str ("?draw_piecewise"), &
!           intrinsic=.true.)
!     call var_list_append_log &
!          (global%var_list, var_str ("?fill_curve"), &
!           intrinsic=.true.)
!     call var_list_append_log &
!          (global%var_list, var_str ("?draw_curve"), &
!           intrinsic=.true.)
!     call var_list_append_log &
!          (global%var_list, var_str ("?draw_errors"), &
!           intrinsic=.true.)
!     call var_list_append_log &
!          (global%var_list, var_str ("?draw_symbols"), &
!           intrinsic=.true.)
!     call var_list_append_string &
!          (global%var_list, var_str ("$fill_options"), &
!           intrinsic=.true.)
!     call var_list_append_string &
!          (global%var_list, var_str ("$draw_options"), &
!           intrinsic=.true.)
!     call var_list_append_string &
!          (global%var_list, var_str ("$err_options"), &
!           intrinsic=.true.)
!     call var_list_append_string &
!          (global%var_list, var_str ("$symbol"), &
!           intrinsic=.true.)
!     call var_list_append_real (global%var_list, &
!          var_str ("tolerance"), 0._default, &
!           intrinsic=.true.)
!     call var_list_append_int (global%var_list, &
!          var_str ("checkpoint"), intrinsic = .true.)
!     call var_list_append_string &
!          (global%var_list, var_str ("$out_file"), var_str (""), &
!           intrinsic=.true.)
!     call var_list_append_log &
!          (global%var_list, var_str ("?out_advance"), .true., &
!           intrinsic=.true.)
!     call var_list_append_log &
!          (global%var_list, var_str ("?out_custom"), .false., &
!           intrinsic=.true.)
!     call var_list_append_string &
!          (global%var_list, var_str ("$out_comment"), var_str ("# "), &
!           intrinsic=.true.)
!     call var_list_append_string &
!          (global%var_list, var_str ("$out_separator"), var_str (" "), &
!           intrinsic=.true.)
!     call var_list_append_log &
!          (global%var_list, var_str ("?out_columns"), .true., &
!           intrinsic=.true.)
!     call var_list_append_log &
!          (global%var_list, var_str ("?out_header"), .true., &
!           intrinsic=.true.)
!     call var_list_append_log &
!          (global%var_list, var_str ("?out_yerr"), .true., &
!           intrinsic=.true.)
!     call var_list_append_log &
!          (global%var_list, var_str ("?out_xerr"), .true., &
!           intrinsic=.true.)
!     call var_list_append_int (global%var_list, var_str ("real_range"), &
!          range (real_specimen), intrinsic = .true., locked = .true.)
!     call var_list_append_int (global%var_list, var_str ("real_precision"), &
!          precision (real_specimen), intrinsic = .true., locked = .true.)
!     call var_list_append_real (global%var_list, var_str ("real_epsilon"), &
!          epsilon (real_specimen), intrinsic = .true., locked = .true.)
!     call var_list_append_real (global%var_list, var_str ("real_tiny"), &
!          tiny (real_specimen), intrinsic = .true., locked = .true.)
!     call var_list_append_log &
!          (global%var_list, var_str ("?polarized_events"), .false., &
!             intrinsic=.true.)
!     ! default settings for shower
!     call var_list_append_log &
!          (global%var_list, var_str ("?ps_fsr_active"), .false., &
!             intrinsic=.true.)
!     call var_list_append_log &
!          (global%var_list, var_str ("?ps_use_PYTHIA_shower"), .true., &
!             intrinsic=.true.)
!     call var_list_append_log &
!          (global%var_list, var_str ("?ps_PYTHIA_verbose"), .false., &
!             intrinsic=.true.)
!     call var_list_append_string &
!          (global%var_list, var_str ("$ps_PYTHIA_PYGIVE"), var_str (""), &
!           intrinsic=.true.)
!     call var_list_append_log &
!          (global%var_list, var_str ("?ps_isr_active"), .false., &
!             intrinsic=.true.)
!     call var_list_append_real (global%var_list, var_str ("ps_mass_cutoff"), &
!          1._default, intrinsic = .true.)
!     call var_list_append_real (global%var_list, var_str ("ps_fsr_lambda"), &
!          0.29_default, intrinsic = .true.)
!     call var_list_append_real (global%var_list, var_str ("ps_isr_lambda"), &
!          0.29_default, intrinsic = .true.)
!     call var_list_append_int (global%var_list, var_str ("ps_max_n_flavors"), &
!          5, intrinsic = .true.)
!     call var_list_append_log &
!          (global%var_list, var_str ("?ps_isr_alpha_s_running"), .true., &
!             intrinsic=.true.)
!     call var_list_append_log &
!          (global%var_list, var_str ("?ps_fsr_alpha_s_running"), .true., &
!             intrinsic=.true.)
!     call var_list_append_real (global%var_list, var_str ("ps_fixed_alpha_s"), &
!          0._default, intrinsic = .true.)
!     call var_list_append_log &
!          (global%var_list, var_str ("?ps_isr_pt_ordered"), .false., &
!             intrinsic=.true.)
!     call var_list_append_log &
!          (global%var_list, var_str ("?ps_isr_angular_ordered"), .true., &
!             intrinsic=.true.)
!     call var_list_append_real (global%var_list, var_str ("ps_isr_primordial_kt_width"), &
!          0._default, intrinsic = .true.)
!     call var_list_append_real (global%var_list, var_str ("ps_isr_primordial_kt_cutoff"), &
!          5._default, intrinsic = .true.)
!     call var_list_append_real (global%var_list, var_str ("ps_isr_z_cutoff"), &
!          0.999_default, intrinsic = .true.)
!     call var_list_append_real (global%var_list, var_str ("ps_isr_minenergy"), &
!          1._default, intrinsic = .true.)
!     call var_list_append_real (global%var_list, var_str ("ps_isr_tscalefactor"), &
!          1._default, intrinsic = .true.)
!     call var_list_append_log &
!          (global%var_list, var_str ("?ps_isr_only_onshell_emitted_partons"), .false., &
!             intrinsic=.true.)
!     ! default settings for hadronization
!     call var_list_append_log &
!          (global%var_list, var_str ("?hadronization_active"), .false., &
!             intrinsic=.true.)
!     ! setting for mlm matching
!     call var_list_append_log &
!          (global%var_list, var_str ("?mlm_matching"), .false., &
!             intrinsic=.true.)
!     call var_list_append_real (global%var_list, var_str ("mlm_Qcut_ME"), &
!          0._default, intrinsic = .true.)
!     call var_list_append_real (global%var_list, var_str ("mlm_Qcut_PS"), &
!          0._default, intrinsic = .true.)
!     call var_list_append_real (global%var_list, var_str ("mlm_ptmin"), &
!          0._default, intrinsic = .true.)
!     call var_list_append_real (global%var_list, var_str ("mlm_etamax"), &
!          0._default, intrinsic = .true.)
!     call var_list_append_real (global%var_list, var_str ("mlm_Rmin"), &
!          0._default, intrinsic = .true.)
!     call var_list_append_real (global%var_list, var_str ("mlm_Emin"), &
!          0._default, intrinsic = .true.)
!     call var_list_append_int (global%var_list, var_str ("mlm_nmaxMEjets"), &
!          0, intrinsic = .true.)
! 
!     call var_list_append_real (global%var_list, var_str ("mlm_ETclusfactor"), &
!          0.2_default, intrinsic = .true.)
!     call var_list_append_real (global%var_list, var_str ("mlm_ETclusminE"), &
!          5._default, intrinsic = .true.)
!     call var_list_append_real (global%var_list, var_str ("mlm_etaclusfactor"), &
!          1._default, intrinsic = .true.)
!     call var_list_append_real (global%var_list, var_str ("mlm_Rclusfactor"), &
!          1._default, intrinsic = .true.)
!     call var_list_append_real (global%var_list, var_str ("mlm_Eclusfactor"), &
!          1._default, intrinsic = .true.)
!     call var_list_append_log &
!          (global%var_list, var_str ("?ckkw_matching"), .false., &
!             intrinsic=.true.)
!     call var_list_append_log &
!          (global%var_list, var_str ("?muli_active"), .false., &
!             intrinsic=.true.)
! 
!     call var_list_append_string (global%var_list, var_str ("$datafile"), &
!           intrinsic=.true.)
!     call var_list_append_string (global%var_list, &
!           var_str ("$comment_prefix"), var_str ("#"), intrinsic=.true.)
!     call var_list_append_log (global%var_list, var_str ("?write_header"), &
!           .true., intrinsic=.true.)
    !!! JRR: WK please check
    !!! Do we really need the variables for pdf_builtin ?
    call var_list_append_string (global%var_list, &
         var_str ("$pdf_builtin_path"), intrinsic=.true.)
    call var_list_append_string (global%var_list, &
         var_str ("$pdf_builtin_set"), intrinsic=.true.)
    call var_list_append_log &
         (global%var_list, var_str ("?omega_openmp"), &
         openmp_is_active (), &
         intrinsic=.true.)
!     call var_list_append_log &
!          (global%var_list, var_str ("?openmp_is_active"), &
!          openmp_is_active (), &
!          locked=.true., intrinsic=.true.)
!     call var_list_append_int &
!          (global%var_list, var_str ("openmp_num_threads_default"), &
!          openmp_get_default_max_threads (), &
!          locked=.true., intrinsic=.true.)
!     call var_list_append_int &
!          (global%var_list, var_str ("openmp_num_threads"), &
!          openmp_get_max_threads (), &
!          intrinsic=.true.)
!     call var_list_append_log &
!          (global%var_list, var_str ("?dipole_resolve_emission"), &
!             intrinsic = .true.)
!     call var_list_append_string &
!          (global%var_list, var_str ("$recombination_mode"), intrinsic = .true.)
!     call var_list_append_log &
!          (global%var_list, var_str ("?recombination_complement"), &
!             intrinsic = .true.)
!     call var_list_append_real &
!          (global%var_list, var_str ("recombination_scale"), intrinsic = .true.)
!     call var_list_append_real &
!          (global%var_list, var_str ("photon_beam_separation"), &
!             intrinsic = .true.)
    call global%init_pointer_variables ()
!    call iterations_lists_init_default (global%it_list_default)
  end subroutine rt_data_global_init                            ! $

  subroutine rt_data_local_init (local, global, env)
    class(rt_data_t), intent(inout), target :: local
    type(rt_data_t), intent(in), target :: global
    integer, intent(in), optional :: env
    call var_list_link (local%var_list, global%var_list)
    if (associated (global%model)) then
       call var_list_init_copies (local%var_list, &
            model_get_var_list_ptr (global%model), &
            derived_only = .true.)
    end if
    call local%init_pointer_variables ()
!    if (present (env)) local%environment = env
  end subroutine rt_data_local_init

  subroutine rt_data_init_pointer_variables (local)
    class(rt_data_t), intent(inout), target :: local
    logical, target, save :: known = .true.
    call var_list_append_string_ptr &
         (local%var_list, var_str ("$fc"), local%os_data%fc, known, &
          intrinsic=.true.)
    call var_list_append_string_ptr &
         (local%var_list, var_str ("$fcflags"), local%os_data%fcflags, known, &
         intrinsic=.true.)
  end subroutine rt_data_init_pointer_variables

  subroutine rt_data_link (local, global)
    class(rt_data_t), intent(inout), target :: local
    type(rt_data_t), intent(in), target :: global
    local%lexer => global%lexer
    call var_list_link (local%var_list, global%var_list)
    if (associated (global%model)) then
       call var_list_synchronize (local%var_list, &
            model_get_var_list_ptr (global%model), reset_pointers = .true.)
    end if
    local%it_list = global%it_list
!    local%it_list_default => global%it_list_default
    local%os_data = global%os_data
    local%model => global%model
    local%prclib_stack = global%prclib_stack
    local%beam_data = global%beam_data
    local%beam_structure = global%beam_structure
    local%process_stack = global%process_stack
!    local%lhapdf_status = global%lhapdf_status
!    local%sf_list_allocated = .false.
!    local%sf_list => global%sf_list
    local%pn = global%pn
    if (allocated (local%sample_fmt))  deallocate (local%sample_fmt)
    if (allocated (global%sample_fmt)) then
       allocate (local%sample_fmt (size (global%sample_fmt)), &
            source = global%sample_fmt)
    end if
!    local%out_files => global%out_files
!    local%rng => global%rng
!    local%beam_polarization => global%beam_polarization
!    local%analysis_data_unit = global%analysis_data_unit
!    call nlo_setup_list_init (local%nlo_setup_list)
  end subroutine rt_data_link

  subroutine rt_data_restore (global, local, keep_model_vars)
    class(rt_data_t), intent(inout) :: global
    type(rt_data_t), intent(inout) :: local
    logical, intent(in), optional :: keep_model_vars
    logical :: same_model, restore
    if (associated (global%model)) then 
       same_model = &
            model_get_name (global%model) == model_get_name (local%model)
       if (present (keep_model_vars) .and. same_model) then
          restore = .not. keep_model_vars
       else
          if (.not. same_model)  call msg_message ("Restoring model '" // &
               char (model_get_name (global%model)) // "'")
          restore = .true.
       end if
       if (restore) then
          call var_list_restore (global%var_list)
       else
          call var_list_synchronize &
               (global%var_list, model_get_var_list_ptr (global%model))
       end if
    end if
    call var_list_undefine (local%var_list, follow_link=.false.)
    global%prclib_stack = local%prclib_stack
    global%process_stack = local%process_stack
!    call nlo_setup_list_final (local%nlo_setup_list)
  end subroutine rt_data_restore

  subroutine rt_data_global_final (global)
    class(rt_data_t), intent(inout) :: global
    call global%process_stack%final ()
    call global%prclib_stack%final ()
    call var_list_final (global%var_list)
!    if (global%sf_list_allocated) then
!       call sf_list_final (global%sf_list)
!       deallocate (global%sf_list)
!       global%sf_list_allocated = .false.
!    end if
!    deallocate (global%it_list_default)
!    call file_list_final (global%out_files)
!    deallocate (global%out_files)
!    deallocate (global%rng)
  end subroutine rt_data_global_final

  subroutine rt_data_add_prclib (global, prclib_entry)
    class(rt_data_t), intent(inout) :: global
    type(prclib_entry_t), intent(inout), pointer :: prclib_entry
    call global%prclib_stack%push (prclib_entry)
    call global%update_prclib (global%prclib_stack%get_first_ptr ())
  end subroutine rt_data_add_prclib
  
  subroutine rt_data_update_prclib (global, lib)
    class(rt_data_t), intent(inout) :: global
    type(process_library_t), intent(in), target :: lib
    type(var_entry_t), pointer :: var
    global%prclib => lib
    var => var_list_get_var_ptr (global%var_list, &
         var_str ("$library_name"), follow_link = .false.)
    if (associated (var)) then
       call var_entry_set_string (var, &
            global%prclib%get_name (), is_known=.true.)
    else
       call var_list_append_string (global%var_list, &
            var_str ("$library_name"), global%prclib%get_name (), &
            intrinsic = .true.)
    end if
  end subroutine rt_data_update_prclib

  function rt_data_get_helicity_selection (rt_data) result (helicity_selection)
    class(rt_data_t), intent(in) :: rt_data
    type(helicity_selection_t) :: helicity_selection
    associate (var_list => rt_data%var_list)
      helicity_selection%active = var_list_get_lval (var_list, &
           var_str ("?helicity_selection_active"))
      if (helicity_selection%active) then
         helicity_selection%threshold = var_list_get_rval (var_list, &
              var_str ("helicity_selection_threshold"))
         helicity_selection%cutoff = var_list_get_ival (var_list, &
              var_str ("helicity_selection_cutoff"))
      end if
    end associate
  end function rt_data_get_helicity_selection

  subroutine write_separator (u)
    integer, intent(in) :: u
    write (u, "(A)")  repeat ("-", 72)
  end subroutine write_separator
  
  subroutine write_separator_double (u)
    integer, intent(in) :: u
    write (u, "(A)")  repeat ("=", 72)
  end subroutine write_separator_double
  

  subroutine rt_data_test (u, results)
    integer, intent(in) :: u
    type(test_results_t), intent(inout) :: results
    call test (rt_data_1, "rt_data_1", &
         "initialize", &
         u, results)
    call test (rt_data_2, "rt_data_2", &
         "fill", &
         u, results)
    call test (rt_data_3, "rt_data_3", &
         "save/restore", &
         u, results)
    call test (rt_data_4, "rt_data_4", &
         "show variables", &
         u, results)
    call test (rt_data_5, "rt_data_5", &
         "show parts", &
         u, results)
end subroutine rt_data_test

  subroutine rt_data_1 (u)
    integer, intent(in) :: u
    type(rt_data_t), target :: rt_data
    
    write (u, "(A)")  "* Test output: rt_data_1"
    write (u, "(A)")  "*   Purpose: initialize global runtime data"
    write (u, "(A)")

    call rt_data%global_init ()
    call rt_data%it_list%init ([2, 3], [5000, 20000])
    call var_list_set_log (rt_data%var_list, var_str ("?omega_openmp"), &
         .false., is_known = .true.)
    rt_data%os_data%fc = "Fortran-compiler"
    rt_data%os_data%fcflags = "Fortran-flags"
    call rt_data%write (u)

    call rt_data%final ()
    
    write (u, "(A)")
    write (u, "(A)")  "* Test output end: rt_data_1"
    
  end subroutine rt_data_1
  
  subroutine rt_data_2 (u)
    integer, intent(in) :: u
    type(rt_data_t), target :: rt_data
    type(var_list_t), pointer :: model_vars
    type(flavor_t), dimension(2) :: flv
    type(string_t) :: cut_expr_text
    type(ifile_t) :: ifile
    type(stream_t) :: stream
    type(parse_tree_t) :: parse_tree
    
    write (u, "(A)")  "* Test output: rt_data_2"
    write (u, "(A)")  "*   Purpose: initialize global runtime data &
         &and fill contents"
    write (u, "(A)")

    call syntax_model_file_init ()

    call rt_data%global_init ()

    call model_list_read_model (var_str ("Test"), &
         var_str ("Test.mdl"), rt_data%os_data, rt_data%model)
    call var_list_set_string (rt_data%var_list, var_str ("$model_name"), &
         model_get_name (rt_data%model), is_known=.true.)
    model_vars => model_get_var_list_ptr (rt_data%model)
    call var_list_init_copies &
         (rt_data%var_list, model_vars)
    call var_list_synchronize &
         (rt_data%var_list, model_vars, reset_pointers = .true.)

    call var_list_set_real (rt_data%var_list, var_str ("sqrts"),&
         1000._default, is_known = .true.)
    call flavor_init (flv, [25,25], rt_data%model)
    call beam_data_init_sqrts (rt_data%beam_data, &
         var_list_get_rval (rt_data%var_list, var_str ("sqrts")), flv)
    
    call var_list_set_string (rt_data%var_list, var_str ("$run_id"), &
         var_str ("run1"), is_known = .true.)
    call var_list_set_real (rt_data%var_list, var_str ("luminosity"), &
         33._default, is_known = .true.)
    
    call syntax_pexpr_init ()
    cut_expr_text = "all Pt > 100 [s]"
    call ifile_append (ifile, cut_expr_text)
    call stream_init (stream, ifile)
    call parse_tree_init_lexpr (parse_tree, stream, .true.)
    rt_data%pn%cuts_lexpr => parse_tree_get_root_ptr (parse_tree)
    
    allocate (rt_data%sample_fmt (2))
    rt_data%sample_fmt(1) = "foo_fmt"
    rt_data%sample_fmt(2) = "bar_fmt"
    
    call var_list_set_log (rt_data%var_list, var_str ("?omega_openmp"), &
         .false., is_known = .true.)
    rt_data%os_data%fc = "Fortran-compiler"
    rt_data%os_data%fcflags = "Fortran-flags"

    call rt_data%write (u)

    call parse_tree_final (parse_tree)
    call stream_final (stream)
    call ifile_final (ifile)
    call syntax_pexpr_final ()

    call rt_data%final ()
    call model_list_final ()
    
    write (u, "(A)")
    write (u, "(A)")  "* Test output end: rt_data_2"
    
  end subroutine rt_data_2
  
  subroutine rt_data_3 (u)
    integer, intent(in) :: u
    type(rt_data_t), target :: rt_data, local
    type(var_list_t), pointer :: model_vars
    type(var_entry_t), pointer :: var
    type(flavor_t), dimension(2) :: flv
    type(string_t) :: cut_expr_text
    type(ifile_t) :: ifile
    type(stream_t) :: stream
    type(parse_tree_t) :: parse_tree
    type(prclib_entry_t), pointer :: lib
    
    write (u, "(A)")  "* Test output: rt_data_3"
    write (u, "(A)")  "*   Purpose: initialize global runtime data &
         &and fill contents;"
    write (u, "(A)")  "*            copy to local block and back"
    write (u, "(A)")

    write (u, "(A)")  "* Init global data"
    write (u, "(A)")

    call syntax_model_file_init ()

    call rt_data%global_init ()

    call model_list_read_model (var_str ("Test"), &
         var_str ("Test.mdl"), rt_data%os_data, rt_data%model)
    call var_list_set_string (rt_data%var_list, var_str ("$model_name"), &
         model_get_name (rt_data%model), is_known=.true.)
    model_vars => model_get_var_list_ptr (rt_data%model)
    call var_list_init_copies &
         (rt_data%var_list, model_vars)
    call var_list_synchronize &
         (rt_data%var_list, model_vars, reset_pointers = .true.)

    call var_list_set_real (rt_data%var_list, var_str ("sqrts"),&
         1000._default, is_known = .true.)
    call flavor_init (flv, [25,25], rt_data%model)
    call beam_data_init_sqrts (rt_data%beam_data, &
         var_list_get_rval (rt_data%var_list, var_str ("sqrts")), flv)
    
    call rt_data%beam_structure%init ([1], rt_data%var_list)
    call rt_data%beam_structure%set_entry (1, 1, var_str ("pdf_builtin"))

    call var_list_set_string (rt_data%var_list, var_str ("$run_id"), &
         var_str ("run1"), is_known = .true.)
    call var_list_set_real (rt_data%var_list, var_str ("luminosity"), &
         33._default, is_known = .true.)
    
    call syntax_pexpr_init ()
    cut_expr_text = "all Pt > 100 [s]"
    call ifile_append (ifile, cut_expr_text)
    call stream_init (stream, ifile)
    call parse_tree_init_lexpr (parse_tree, stream, .true.)
    rt_data%pn%cuts_lexpr => parse_tree_get_root_ptr (parse_tree)
    
    allocate (rt_data%sample_fmt (2))
    rt_data%sample_fmt(1) = "foo_fmt"
    rt_data%sample_fmt(2) = "bar_fmt"

    call var_list_set_log (rt_data%var_list, var_str ("?omega_openmp"), &
         .false., is_known = .true.)
    rt_data%os_data%fc = "Fortran-compiler"
    rt_data%os_data%fcflags = "Fortran-flags"
    
    allocate (lib)
    call lib%init (var_str ("library_1"))
    call rt_data%add_prclib (lib)

    write (u, "(A)")  "* Init and modify local data"
    write (u, "(A)")

    call local%local_init (rt_data)
    call local%link (rt_data)

    var => var_list_get_var_ptr (local%var_list, var_str ("ms"))
    if (var_entry_is_copy (var)) then
       call var_list_init_copy (local%var_list, var, user=.true.)
       call var_list_set_original_pointer (local%var_list, var_str ("ms"), &
            model_vars)
       call var_list_set_real (local%var_list, var_str ("ms"), &
         150._default, is_known = .true., model_name = var_str ("Test"))
       call model_parameters_update (local%model)
       call var_list_synchronize (local%var_list, model_vars)
    end if

    call var_list_append_string (local%var_list, &
         var_str ("$integration_method"), intrinsic = .true., user = .true.)
    call var_list_set_string (local%var_list, var_str ("$integration_method"), &
         var_str ("midpoint"), is_known = .true.)
    
    call var_list_append_string (local%var_list, &
         var_str ("$phs_method"), intrinsic = .true., user = .true.)
    call var_list_set_string (local%var_list, var_str ("$phs_method"), &
         var_str ("single"), is_known = .true.)

    local%os_data%fc = "Local compiler"
    
    allocate (lib)
    call lib%init (var_str ("library_2"))
    call local%add_prclib (lib)

    call local%write (u)

    write (u, "(A)")
    write (u, "(A)")  "* Restore global data"
    write (u, "(A)")
    
    call rt_data%restore (local)
    call rt_data%write (u)

    write (u, "(A)")
    write (u, "(A)")  "* Cleanup"
    
    call parse_tree_final (parse_tree)
    call stream_final (stream)
    call ifile_final (ifile)
    call syntax_pexpr_final ()

    call rt_data%final ()
    call model_list_final ()
    
    write (u, "(A)")
    write (u, "(A)")  "* Test output end: rt_data_3"
    
  end subroutine rt_data_3
  
  subroutine rt_data_4 (u)
    integer, intent(in) :: u
    type(rt_data_t), target :: rt_data
    
    type(string_t), dimension(0) :: empty_string_array

    write (u, "(A)")  "* Test output: rt_data_4"
    write (u, "(A)")  "*   Purpose: display selected variables"
    write (u, "(A)")

    call rt_data%global_init ()

    write (u, "(A)")  "* No variables:"
    write (u, "(A)")

    call rt_data%write_vars (u, empty_string_array)

    write (u, "(A)")  "* Two variables:"
    write (u, "(A)")

    call rt_data%write_vars (u, &
         [var_str ("?unweighted"), var_str ("$phs_method")])
    
    write (u, "(A)")
    write (u, "(A)")  "* Display whole record with selected variables"
    write (u, "(A)")

    call rt_data%write (u, &
         vars = [var_str ("?unweighted"), var_str ("$phs_method")])
    
    write (u, "(A)")
    write (u, "(A)")  "* Test output end: rt_data_4"
    
  end subroutine rt_data_4
  
  subroutine rt_data_5 (u)
    integer, intent(in) :: u
    type(rt_data_t), target :: rt_data
    
    write (u, "(A)")  "* Test output: rt_data_5"
    write (u, "(A)")  "*   Purpose: display parts of rt data"
    write (u, "(A)")

    call rt_data%global_init ()
    call rt_data%write_libraries (u)

    write (u, "(A)")

    call rt_data%write_beams (u)

    write (u, "(A)")

    call rt_data%write_process_stack (u)

    call rt_data%final ()
    
    write (u, "(A)")
    write (u, "(A)")  "* Test output end: rt_data_5"
    
  end subroutine rt_data_5
  

end module rt_data
